<?php

/**
 * @file
 * Contains \Drupal\views_field_compare\Plugin\views\filter\FieldContained.
 */

namespace Drupal\views_field_compare\Plugin\views\filter;

use Drupal\Core\Form\FormStateInterface;
use Drupal\views\Plugin\views\field\EntityField;
use Drupal\views\Plugin\views\filter\FilterPluginBase;

/**
 * Filter handler which allows to filter by a field value contained in or not
 * contained in another multi-valued field.
 *
 * @ingroup views_field_handlers
 *
 * @ViewsFilter("field_contained")
 */
class FieldContained extends FilterPluginBase {

  /**
   * {@inheritdoc}
   */
  public function canExpose() {
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  protected function canBuildGroup() {
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    $options = parent::defineOptions();
    $options['left_field'] = ['default' => ''];
    $options['right_field'] = ['default' => ''];
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function operatorOptions($which = 'title') {
    $options = [];
    foreach ($this->operators() as $id => $info) {
      $options[$id] = $info[$which];
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildOptionsForm($form, $form_state);
    $this->view->initStyle();

    // Allow to choose all fields as possible
    if ($this->view->style_plugin->usesFields()) {
      $left_options = [];
      $right_options = [];
      /** @var \Drupal\views\Plugin\views\field\EntityField $field */
      foreach ($this->view->display_handler->getHandlers('field') as $name => $field) {
        // Only allow clickSortable fields for left options.
        if ($field->clickSortable()) {
          $left_options[$name] = $field->adminLabel(TRUE);
        }
        elseif ($field->multiple) {
          // Only allow multi-valued fields for right options.
          $right_options[$name] = $field->adminLabel(TRUE);
        }
      }
      if (!empty($left_options)) {
        $form['left_field'] = [
          '#type' => 'select',
          '#title' => $this->t('Choose left field'),
          '#description' => $this->t("This filter doesn't work for very special field handlers."),
          '#options' => $left_options,
          '#default_value' => $this->options['left_field'],
        ];
      }
      if (!empty($right_options)) {
        $form['right_field'] = [
          '#type' => 'select',
          '#title' => $this->t('Choose right field'),
          '#description' => $this->t("This filter requires multi-valued field handlers."),
          '#options' => $right_options,
          '#default_value' => $this->options['right_field'],
        ];
      }
      if (empty($left_options) || empty($right_options)) {
        $form_state->setErrorByName('', $this->t('You have to add some fields to be able to use this filter.'));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function validate() {
    $errors = parent::validate();
    $fields = $this->displayHandler->getHandlers('field');
    if ($this->displayHandler->usesFields()) {
      foreach ([
                 $this->options['left_field'],
                 $this->options['right_field'],
               ] as $id) {
        if (!isset($fields[$id])) {
          // Combined field filter only works with fields that are in the fields list.
          $errors[] = $this->t('Field %field set in %filter is not set in this display.', [
            '%field' => $id,
            '%filter' => $this->adminLabel(),
          ]);
          break;
        }
      }

      if (!empty($this->options['left_field'])
        && isset($fields[$this->options['left_field']])
        && !$fields[$this->options['left_field']]->clickSortable()) {
        // Combined field filter only works with simple fields. If the field is
        // not click sortable we can assume it is not a simple field.
        // @todo change this check to isComputed. See
        // https://www.drupal.org/node/2349465
        $errors[] = $this->t('Field %field set in %filter is not usable for this filter type. Fields contained filter only works for simple fields.', [
          '%field' => $fields[$id]->adminLabel(),
          '%filter' => $this->adminLabel(),
        ]);
      }

      //$left_entity_type = $fields[$this->options['left_field']]->definition['entity_type'];
      //$right_entity_type = $fields[$this->options['right_field']]->definition['entity_type'];
      //if ($left_entity_type != $right_entity_type) {
      // Both fields must be attached to the same entity type for the subquery expression
      // used in the query function to work correctly.
      // @TODO enhance the subquery expression to allow fields from different entity types
      //  $errors[] = $this->t('Left and right fields must be attached to same entity type. Left entity = %left, right entity = %right', [
      //    '%left' => $left_entity_type,
      //    '%right' => $right_entity_type,
      //  ]);
      //}

    }
    else {
      $errors[] = $this->t('%display: %filter can only be used on displays that use fields.', [
        '%display' => $this->displayHandler->display['display_title'],
        '%filter' => $this->adminLabel()
      ]);
    }
    return $errors;
  }

  /**
   * {@inheritdoc}
   */
  function operators() {
    $operators = [
      'in' => [
        'title' => $this->t('Is contained in'),
        'short' => $this->t('in'),
        'short_single' => $this->t('IN'),
        'method' => 'opContained',
        'values' => 1,
      ],
      'not in' => [
        'title' => $this->t('Is not contained in'),
        'short' => $this->t('not in'),
        'short_single' => $this->t('!IN'),
        'method' => 'opNotContained',
        'values' => 1,
      ],
    ];
    return $operators;
  }

  /**
   * {@inheritdoc}
   */
  public function query() {
    $this->view->_build('field');
    $fields = [];
    // Only add the fields if they have a proper field and table alias.
    foreach ([
      $this->options['left_field'],
      $this->options['right_field'],
    ] as $id) {
      // Overridden fields can lead to fields missing from a display that are
      // still set in the non-overridden combined filter.
      if (!isset($this->view->field[$id])) {
        // If fields are no longer available that are needed to filter by, make
        // sure no results are shown to prevent displaying more then intended.
        $this->view->build_info['fail'] = TRUE;
        continue;
      }
      /** @var \Drupal\views\Plugin\views\field\EntityField $field */
      $field = $this->view->field[$id];
      // Always add the table of the selected fields to be sure a table alias exists.
      $field->ensureMyTable();
      if (!empty($field->tableAlias) && !empty($field->realField)) {
        $fields[] = $field;
      }
    }

    if ($fields && count($fields) == 2) {
      $info = $this->operators();
      if (!empty($info[$this->operator]['method'])) {
        $this->{$info[$this->operator]['method']}($fields[0], $fields[1]);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function opContained($source_field, $target_field) {
    $subquery = $this->generateSubquery($source_field, $target_field);
    if ($subquery) {
      $expression = "$source_field->tableAlias.$source_field->realField" . " IN (\n" . $subquery . "\n)";
      $this->query->addWhereExpression($this->options['group'], $expression);
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function opNotContained($source_field, $target_field) {
    $subquery = $this->generateSubquery($source_field, $target_field);
    if ($subquery) {
      $expression = "$source_field->tableAlias.$source_field->realField" . " NOT IN (\n" . $subquery . "\n)";
      $this->query->addWhereExpression($this->options['group'], $expression);
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function generateSubquery(EntityField $source_field, EntityField $target_field) {
    $source_table = $source_field->table;
    $source_table_alias = $this->tableAlias($source_table);
    $target_table = $target_field->table;
    $target_table_alias = $this->tableAlias($target_table);
    $target_column = $target_field->realField;

    /** @var \Drupal\views\Plugin\views\query\Sql $query */
    $query = $target_field->query;
    $tableQueue = $query->getTableQueue();
    $joins = [];
    $target = TRUE;
    foreach ([$target_table, $source_table] as $relationship) {
      while ($relationship) {
        foreach ($tableQueue as $table_alias => $table_data) {
          if ($table_alias == $relationship || $table_data['table'] == $relationship) {
            if ($relationship == $source_table) {
              $source_table_data = $table_data;
            }
            if (isset($table_data['join'])) {
              /** @var \Drupal\views\Plugin\views\join\Standard $join */
              $join = $table_data['join'];
              // Find the real name of the left table for the join
              $left_table = $tableQueue[$join->leftTable]['table'];
              $left_table_alias = $this->tableAlias($left_table);
              // Determine which table to join depending on whether processing source or target fields.
              $join_table = $target ? $left_table : $join->table;
              $join_table_alias = $this->tableAlias($join_table);
              $right_table_alias = $this->tableAlias($join->table);
              $join_sql = "$join->type JOIN [$join_table] $join_table_alias ON $left_table_alias.$join->leftField = $right_table_alias.$join->field";
              if ($join->extra) {
                foreach ($join->extra as $extra) {
                  $join_sql .= " {$join->extraOperator} {$right_table_alias}.{$extra['field']} = {$extra['value']}";
                }
              }
              $joins[] = $join_sql;
              $relationship = isset($table_data['relationship']) ? $table_data['relationship'] : FALSE;
            }
            else {
              $relationship = ($relationship == $left_table) ? FALSE : $left_table;
            }
            break;
          }
        }
      }

      // Save last right table details
      $right_tables[] = "$right_table_alias.$join->field";

      // Finished processing for $target_field, next iteration will process $source_field
      if ($target) {
        $target_joins = count($joins);
        $target = FALSE;
      }
    }

    if ($joins) {
      // Reverse the order of the joins from the source field and generate a string.
      $joins = array_merge(array_slice($joins, 0, $target_joins,), array_reverse(array_slice($joins, $target_joins,)));

      // Generate the subquery sql components
      $sql[] = "SELECT $target_column FROM [$target_table] $target_table_alias";
      $sql[] = implode("\n", $joins);
      $sql[] = "WHERE $right_tables[0] = $right_tables[1]";
      $sql[] = "AND {$source_table_alias}.{$source_table_data['join']->field} = {$source_table_data['alias']}.{$source_table_data['join']->field}";
      // Replace the square brackets around table names with curly braces to conform to Drupal standard.
      return str_replace(['[', ']'], ['{', '}'], implode("\n", $sql));
    }

    return FALSE;
  }

  protected function tableAlias($table) {
    // Generate a table alias by stripping all underscore chars.
    return str_replace('_', '', $table);
  }

}
